#include <Windows.h>

#include <shobjidl.h>
#include <propkey.h>
#include <propvarutil.h>

#include <iostream>
#include <string>
#include <vector>
#include <map>

using namespace std;

namespace JumpListUtils
{
	HRESULT CreateShellLink(PCWSTR szAppPath, PCWSTR pszArguments, PCWSTR pszTitle, IShellLink **ppsl)
	{
		IShellLink *psl = NULL;
		HRESULT hr = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&psl));
		if (SUCCEEDED(hr))
		{
			hr = psl->SetPath(szAppPath);
			if (SUCCEEDED(hr))
			{
				hr = psl->SetArguments(pszArguments);
				if (SUCCEEDED(hr))
				{
					// The title property is required on Jump List items provided as an IShellLink
					// instance.  This value is used as the display name in the Jump List.
					IPropertyStore *pps = NULL;
					hr = psl->QueryInterface(IID_PPV_ARGS(&pps));
					if (SUCCEEDED(hr))
					{
						PROPVARIANT propvar;
						hr = InitPropVariantFromString(pszTitle, &propvar);
						if (SUCCEEDED(hr))
						{
							hr = pps->SetValue(PKEY_Title, propvar);
							if (SUCCEEDED(hr))
							{
								hr = pps->Commit();
								if (SUCCEEDED(hr))
								{
									hr = psl->QueryInterface(IID_PPV_ARGS(ppsl));
								}
							}
							PropVariantClear(&propvar);
						}
						pps->Release();
					}
				}
			}
			psl->Release();
		}
		return hr;
	}

	HRESULT CreateSeparatorLink(IShellLink **ppsl)
	{
		IPropertyStore *pps = NULL;
		HRESULT hr = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&pps));
		if (SUCCEEDED(hr))
		{
			PROPVARIANT propvar;
			hr = InitPropVariantFromBoolean(TRUE, &propvar);
			if (SUCCEEDED(hr))
			{
				hr = pps->SetValue(PKEY_AppUserModel_IsDestListSeparator, propvar);
				if (SUCCEEDED(hr))
				{
					hr = pps->Commit();
					if (SUCCEEDED(hr))
					{
						hr = pps->QueryInterface(IID_PPV_ARGS(ppsl));
					}
				}
				PropVariantClear(&propvar);
			}
			pps->Release();
		}
		return hr;
	}

	class CategoryCollection
	{
	public:
		CategoryCollection(LPCWSTR appId)
		{
			HRESULT hr = CoCreateInstance(CLSID_DestinationList, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&this->dstList));
			if (SUCCEEDED(hr))
			{
				hr = this->dstList->SetAppID(appId);
				if (SUCCEEDED(hr))
				{
					UINT minSlots = 0;
					IObjectArray *removed = NULL;
					hr = this->dstList->BeginList(&minSlots, IID_PPV_ARGS(&removed));
					if (SUCCEEDED(hr))
					{
						removed->Release();
					}
				}
			}
		}

		~CategoryCollection()
		{
			Commit();

			for (NameToCollection::iterator itr = this->catCollection.begin();
				itr != this->catCollection.end();
				++itr)
			{
				if (itr->second)
				{
					itr->second->Release();
					itr->second = NULL;
				}
			}

			if (this->dstList)
			{
				this->dstList->Release();
				this->dstList = NULL;
			}
		}

		void Add(const std::wstring &name, IShellLinkW *item)
		{
			IObjectCollection *collection = this->catCollection[name];
			if (!collection)
			{
				CoCreateInstance(CLSID_EnumerableObjectCollection, NULL, CLSCTX_INPROC, IID_PPV_ARGS(&collection));
			}
			if (collection)
			{
				collection->AddObject(item);
				this->catCollection[name] = collection;
			}
		}

	private:
		void Commit()
		{
			for (NameToCollection::iterator itr = this->catCollection.begin();
				itr != this->catCollection.end();
				++itr)
			{
				IObjectArray *poa = NULL;
				HRESULT hr = itr->second->QueryInterface(IID_PPV_ARGS(&poa));
				if (SUCCEEDED(hr))
				{
					if (itr->first.length() == 1 && 0 == wcscmp(L"-", itr->first.c_str()))
					{
						hr = dstList->AddUserTasks(poa);
					}
					else
					{
						hr = dstList->AppendCategory(itr->first.c_str(), poa);
					}
				
					poa->Release();
				}
			}
			this->dstList->CommitList();
		}

		typedef map<const std::wstring, IObjectCollection*> NameToCollection;
		NameToCollection catCollection;
		ICustomDestinationList *dstList;
	};
}

namespace StrUtils
{
	bool strMatch(const wstring &left, const wstring &right, size_t length)
	{
		return 0 == wcsncmp(left.c_str(), right.c_str(), length);
	}

	bool cmdMatch(const wstring &cmd, const wstring &cmp)
	{
		return cmp.size() >= cmd.size() && strMatch(cmd, cmp, cmd.length());
	}

	// From http://www.codeproject.com/KB/stl/Split_string.aspx
	template <typename E, typename C>
	size_t strsplit(
		std::basic_string<E> const& s,
		C &container,
		E const delimiter,
		bool keepBlankFields = false)
	{
		size_t n = 0;
		std::basic_string<E>::const_iterator it = s.begin(), end = s.end(), first;
		for (first = it; it != end; ++it)
		{
			if (delimiter == *it)
			{
				if (keepBlankFields || first != it)
				{
					container.push_back(std::basic_string<E>(first, it));
					++n;

					first = it + 1;
				}
				else
				{
					++first;
				}
			}
		}
		if (keepBlankFields || first != it)
		{
			container.push_back(std::basic_string<E>(first, it));
			++n;
		}
		return n;
	}
}

int wmain(int argCount, wchar_t *args[])
{
	if (argCount != 2)
	{
		wcout << L"Usage:\n"
			<< L"\tSetJumpList [AppUserModelId]" << endl
			<< endl
			<< L"After starting SetJumpList standard input determines the jump list for the specified AppUserModelId." << endl
			<< L"Each line of standard input defines an entry in the jump list:" << endl
			<< L"\titem [category name],[executable path],[arguments],[name]" << endl
			<< L"\tseparator [category name]" << endl
			<< L"\texit" << endl
			<< endl
			<< L"Use 'item' to add a new entry to 'category' with name 'name' and executing 'executable path' with 'arguments' passed to it." << endl
			<< L"Use 'separator' to add a separator to 'category'." << endl
			<< L"A category name of '-' adds to the User Items list." << endl
			<< L"Use 'exit' to quit and save the jump list." << endl
			<< endl
			<< L"Example: " << endl
			<< L"\tSetJumpList Contoso.GM.Readme.1" << endl
			<< L"\t\titem Readme List,C:\\windows\\notepad.exe,C:\\windows\\system32\\drivers\\gmreadme.txt,GM Readme" << endl
			<< L"\t\tseparator Readme List" << endl
			<< L"\t\titem Readme List,C:\\windows\\notepad.exe,C:\\foo.txt,Foo Text" << endl
			<< L"\t\titem -,C:\\windows\\notepad.exe,C:\\bar.txt,User Task Bar" << endl
			<< L"\t\texit" << endl
			;
	}
	else
	{
		HRESULT hr = CoInitialize(NULL);
		if (SUCCEEDED(hr))
		{
			{
				JumpListUtils::CategoryCollection col(args[1]);
				wstring line;
				bool keepGoing = true;
				const wstring itemCmd(L"item ");
				const wstring sepCmd(L"separator ");
				const wstring exitCmd(L"exit");

				while (keepGoing && getline(wcin, line))
				{
					if (StrUtils::cmdMatch(itemCmd, line))
					{
						vector<wstring> lineParts;
						StrUtils::strsplit(line.substr(itemCmd.length()), lineParts, L',');

						if (lineParts.size() != 4)
						{
							wcout << L"item [category name],[executable path],[arguments],[name]" << endl;
						}
						else
						{
							IShellLinkW *item = NULL;
							hr = JumpListUtils::CreateShellLink(lineParts[1].c_str(), lineParts[2].c_str(), lineParts[3].c_str(), &item);
							if (SUCCEEDED(hr))
							{
								col.Add(lineParts[0], item);
								item->Release();
								item = NULL;
							}
						}
					}
					else if (StrUtils::cmdMatch(sepCmd, line))
					{
						wstring name = line.substr(sepCmd.length());

						if (name.length() == 0)
						{
							wcout << L"separator [category name]" << endl;
						}
						else
						{
							IShellLinkW *item = NULL;
							hr = JumpListUtils::CreateSeparatorLink(&item);
							if (SUCCEEDED(hr))
							{
								col.Add(name, item);
								item->Release();
								item = NULL;
							}
						}
					}
					else if (StrUtils::cmdMatch(exitCmd, line))
					{
						keepGoing = false;
					}
					else
					{
						wcout << L"Unknown command: " << line << endl;

						wcout << L"Instead use: " << endl
							<< L"\titem [category name],[executable path],[arguments],[name]" << endl
							<< L"\tseparator [category name]" << endl
							<< L"\texit" << endl;
					}
				}
			}
			CoUninitialize();
		}
	}
	return 0;
}